/** @file fwUnitTestComponentTestDiscovery.ctl
  
 @brief Automatically finds the test suites and test cases. 

@par Creation Date
		01/04/2014 
  
 */

#uses "fwInstallation.ctl"  // We call fwInstallation_getComponentInfo( )
 
#uses "FwUnitTestComponent/fwUnitTestComponentRemoveComments.ctl" 

const dyn_string TEST_DIRECTORIES = makeDynString("/scripts/test/", "/scripts/libs/test/", "/scripts/libs/Test/", "/scripts/Test/");

const string FWUNIT_TEST_STANDARD_MODE = "Standard";
const string FWUNIT_TEST_INTERACTIVE_MODE = "Interactive";
const string FWUNIT_TEST_BACKGROUND_MODE = "Background";

//------------------------------------------------------------------------------------------------

/** 
  @par Description: Gets the test scripts and test routines for the given component.
  
  @par Usage
	  Public

  @param[in]  componentName string, The component the test script belongs to.
  @param[out] scriptsTestRoutines mapping, key = script path name, value = string[] of test routines.
  @param[in]  sSelectionMode Test selection mode (Standard, Interactive, Background, combined by "|")
*/
void fwUnitTestComponentTestDiscovery_testDiscoveryForComponent(const string componentName, mapping &scriptsTestRoutines, string sSelectionMode="Standard|Background")
{
  mapping testSuiteFiles;
  fwUnitTestComponentTestDiscovery_getTestSuiteFiles(makeDynString(componentName), testSuiteFiles);
  //DebugTN("testSuiteFiles = ",testSuiteFiles);
  fwUnitTestComponentTestDiscovery_getTestCaseRoutines(testSuiteFiles[componentName], scriptsTestRoutines, sSelectionMode);
  //DebugTN("scriptsTestRoutines = ",scriptsTestRoutines);  
}

//------------------------------------------------------------------------------------------------

/** 
  @par Description: In the XML file the path to the test scripts are not present. So we add the path to the
                      test scripts.
  
  @par Usage
	  Public

  @param[in]  componentName string, The component the test script belongs to.
  @param[in]  scriptName string, the name of the test script you wish to find the path of.
 
  @return testScriptPathName string
*/
string fwUnitTestComponentTestDiscovery_getTestScriptPathName(const string componentName, string scriptName)
{  
  //TODO for v2 we said that we would allow the developer to specify different paths for the tests, so
  //   they would do that in the XML file, and here we should check if they have given the path. 
  // on the panel have a file selector option.
  
  mapping testSuiteFiles;
  fwUnitTestComponentTestDiscovery_getTestSuiteFiles(makeDynString(componentName), testSuiteFiles);
  dyn_string testScripts = testSuiteFiles[componentName];
      
  for (int i = 1; i <= dynlen(testScripts); i++)
  {
    if (strtolower(baseName(testScripts[i])) == strtolower(baseName(scriptName)))
    {      
      return testScripts[i];
    }
  }  
  return "";
}

//------------------------------------------------------------------------------------------------

/** 
  @par Description: Get all the tests within the project(including within the installationDirectoryPath).
  
  @par Usage
	  Public

  @param[out] tests mapping, key= component name, value=(mapping key = suitesToRun, value = test cases) 
  @param[in]  sSelectionMode Test selection mode (Standard, Interactive, Background, combined by "|")
*/
void fwUnitTestComponentTestDiscovery_getTests(mapping &tests, string sSelectionMode="Standard|Background")
{
  string componentPath;
  
  // 140923 Decl added by PCB:
  dyn_string componentNames; // Initially empty
    
  
  dyn_string paths;
  fwInstallation_getProjPaths(paths);
  for (int j = 1; j <= dynlen(paths); j++) {
    for (int i = 1; i <= dynlen(TEST_DIRECTORIES); i++) {
      componentPath = paths[j] + TEST_DIRECTORIES[i];
      componentNames = fwUnitTestComponentTestDiscovery_appendArraysNoDuplicates(componentNames, getFileNames(componentPath, "*", FILTER_DIRS));
    }
  }
  
  // get the tests for all the components
  for(int componentIndex = 1; componentIndex <= dynlen(componentNames); componentIndex++)
  { string componentName = componentNames[componentIndex];
    if ((componentName == ".") || (componentName == ".."))
    {
      continue;
      //dynRemove(componentNames, componentIndex);
    }
    else
    {
      mapping scriptsTestRoutines;
      
      

      // The fw Component Installation tool has it's own test routines.
      // As of May 2016, these are being installed alongside the FwInst.
      // fwUnitTestingComponent (ie us) find this FwInst "component"'s test files.
      // We look for the FwInst "component"'s corresponding DPE in the FwInst internal structures.
      // But there is no such DPE (as FwInst does not talk about itself.)
      // So we have to make a special case when the component is called "fwInstallation" :-(
      string componentNameToSkip = "fwInstallation"; // MAGIC NUMBER!

      if ( componentName == componentNameToSkip )     
         { DebugTN("INFO: The Automatic Test Discovery always SKIPS ",componentNameToSkip,"'s own test routines!");
           // We not only skip this entry, but we remove it from the list, so that subsequent code doesn't get confused either!!!  
  
           // NB This code only works because CTRL re-evaluates the dynlen(componentNames) every time around the loop!    
           dynRemove(componentNames,componentIndex);
           //DebugTN("Now componentNames = ",componentNames);
           //DebugTN("Now dynlen(componentNames) = ", dynlen(componentNames) );      
           continue; 
         }
      // So automatic discovery doesn't even offer him the fwInstallation test routines!    

      //DebugTN("testDiscoveryForComponent ",componentName);
      fwUnitTestComponentTestDiscovery_testDiscoveryForComponent(componentName, scriptsTestRoutines, sSelectionMode);
      tests[componentName] = scriptsTestRoutines;
    }
  }
}

//------------------------------------------------------------------------------------------------

/** 
  @par Description: Append array2 to array1 without adding duplicates. (union of two arrays)
         When given a file path name the name of the files are compared(without the path). Priority is given to array1.
  
  @par Usage
	  Public

  @param[in] array1 dyn_string, Array to be appended to.
  @param[in] array2 dyn_string, Array to be appended.
  
  @return dyn_string The result of appending the arrays.  
*/
dyn_string fwUnitTestComponentTestDiscovery_appendArraysNoDuplicates(dyn_string array1, dyn_string array2)
{
  bool found;
  for (int i = 1; i <= dynlen(array2); i++)
  {
    found = false;
    for (int j = 1; j <= dynlen(array1); j++)
    {
      if (baseName(array2[i]) == baseName(array1[j]))
      {
        found = true;
        break;
      }
    }
    if (!found)
    {
      dynAppend(array1, array2[i]);
    }
  }
  return array1;
}


//------------------------------------------------------------------------------------------------
/////////////////////////////// test suite script discovery //////////////////////////////////////////////

/** 
  @par Description: Gets the list of test script files. If a test script is found both within the project path and
                    within the installation directory then priority is given to the version in the
                    project path. 
  
  @par Usage
	  Public

  @param[in]  componentsToTest dyn_string, the name of the components you wish to find the tests for.
  @param[out] testSuiteFiles   mapping, the key is the name of the component, 
                                        value is dyn_string of pathnames for the test suite files
*/
void fwUnitTestComponentTestDiscovery_getTestSuiteFiles(dyn_string componentsToTest, mapping &testSuiteFiles)
{
  string testScriptsPath, testLibsPath; // root directory of the tests for a given component 
  string installationTestScriptDirPath, installationTestLibDirPath;
  string installationTestDirPath; // the test directory within the componentsInstallationDir
  dyn_string testFiles; // list of tests for the current component. 
  bool found = false; // if the test script has already been added to the mapping.
  
  // tests from the project path
  for (int componentIndex = 1; componentIndex <= dynlen(componentsToTest); componentIndex++)
  {
    string componentName = componentsToTest[componentIndex];  
    fwUnitTestComponentTestDiscovery_getComponentTestSuiteFiles(componentName, testSuiteFiles);
  } 
}

//------------------------------------------------------------------------------------------------
/**
 * Gets the test suite files of a component.
 * @param componentName - [IN] The component name.
 * @param tetSuiteFiles - [OUT] Mapping containing the test suite files for the component.
 */
void fwUnitTestComponentTestDiscovery_getComponentTestSuiteFiles(string componentName, mapping & testSuiteFiles) {

  dyn_string paths;
  fwInstallation_getProjPaths(paths);
  testSuiteFiles[componentName] = makeDynString();
  
  // Process the project paths in reverse order (same priority as WinCC to use panels and scripts)
  int iLen = dynlen(paths);
  for(int i=iLen; i>=1; i--) {
    string sPath = paths[i];
    testSuiteFiles[componentName] = ( 
          fwUnitTestComponentTestDiscovery_appendArraysNoDuplicates( 
                    testSuiteFiles[componentName], 
                    fwUnitTestComponentTestDiscovery_getScriptsForComponentsPath(componentName, sPath)
           )
        ); 
  }
}
//------------------------------------------------------------------------------------------------
/** 
  @par Description: Gets the list of test script files for the given component within the given path.
                       
  @see getTestSuiteFiles()
  
  @par Usage
	  Public

  @param[in] componentName string, the name of the component you wish to find the tests for.
  @param[in] path          string, will either be the PROJ_PATH or installationDirectoryPath
*/
private dyn_string fwUnitTestComponentTestDiscovery_getScriptsForComponentsPath(const string componentName, const string path)
{
  dyn_string testFiles = makeDynString();
  for (int i = 1; i <= dynlen(TEST_DIRECTORIES); i++) 
  {  
    dyn_string scripts = makeDynString();  
    if (isdir(path + TEST_DIRECTORIES[i] + componentName))
    {     
       fwUnitTestComponentTestDiscovery_getTestScriptsFrom(path + TEST_DIRECTORIES[i] + componentName, testFiles);
    }
  }
  //DebugTN("testFiles : ", testFiles);
  return testFiles;
}

//------------------------------------------------------------------------------------------------

/** 
  @par Description: Gets the tests scripts from the test path.
  
  @par Usage
	  Public

  @param[in]  testPath string, Path to look in for the test scripts.
  @param[out] testFiles   dyn_string, list of test files within the testPath
*/
void fwUnitTestComponentTestDiscovery_getTestScriptsFrom(const string testPath, dyn_string &testFiles)
{
  fwUnitTestComponentTestDiscovery_getTestScriptsRecurse(testPath, makeDynString(".", ".."), testFiles); 
}

//------------------------------------------------------------------------------------------------

/** 
  @par Description: @see fwUnitTestComponentTestDiscovery_getTestScriptsFrom()
  
  @par Usage
	  Public

  @param[in]  testPath string, Path to look in for the test scripts.
  @param[in]  componentsToExclude dyn_string, Paths to exclude, used to stop multiple calls with the same filePath.
  @param[out] testFiles   dyn_string, list of test files within the testPath
*/
private void fwUnitTestComponentTestDiscovery_getTestScriptsRecurse(const string testPath, dyn_string componentsToExclude, dyn_string &testFiles)
{ 
  
  dyn_string subFiles = getFileNames(testPath , "test*.ctl", FILTER_FILES);
  dyn_string subDirs = getFileNames(testPath, "*", FILTER_DIRS);
  
  if(dynContains(componentsToExclude, ".") <= 0) {
    dynAppend(componentsToExclude, ".");
  }
  
  if(dynContains(componentsToExclude, "..") <= 0) {
    dynAppend(componentsToExclude, "..");    
  }
    
  // append all file paths to the testFiles array  
  for(int j=1 ; j <= dynlen(subFiles) ; j++)
	 {
    if(doesNotAlreadyContainTestScript(testFiles, testPath + "/" + subFiles[j]))
    {
      dynAppend(testFiles, testPath + "/" + subFiles[j]);
    }
	 }	  
  
  //DebugTN(testFiles);  
  // recurse through the directories yet to be explored
  for(int j=1 ; j <= dynlen(subDirs) ; j++)
	 {
    if(dynContains(componentsToExclude, subDirs[j]))
    {  
      continue;
    }  
    fwUnitTestComponentTestDiscovery_getTestScriptsRecurse(testPath + "/" + subDirs[j], componentsToExclude, testFiles);
	 }	   
}

//------------------------------------------------------------------------------------------------

/** 
  @par Description: Used to stop both "/scripts/libs/test/" and "/scripts/libs/Test/" paths for a test being added when using MS Windows.
  
  @par Usage
	  Public

  @param[in]  toBeAddedTestFile string
  @param[out] testFiles   dyn_string, list of test files within the testPath
*/
private bool doesNotAlreadyContainTestScript(dyn_string testFiles, string toBeAddedTestScript)
{
  //DebugTN(testFiles);
  //DebugTN(toBeAddedTestScript);
  for (int i=1 ; i <= dynlen(testFiles) ; i++)
  {
    if(strtolower(testFiles[i]) == strtolower(toBeAddedTestScript))
    {
      return FALSE;
    }
  }
  return TRUE;
}

//------------------------------------------------------------------------------------------------

/////////////////////////////// test routine discovery //////////////////////////////////////////////

/** 
  @par Description: For each of the given testScriptPathNames finds the test case routines it contains. 
  
  @par Usage
	  Public

  @param[in]  testScriptPathNames dyn_string, List of all the test script path names
  @param[out] scriptsTestRoutines mapping, key = testScriptPathName, value = test case routines
  @param[in]  sSelectionMode Test selection mode (Standard, Interactive, Background, combined by "|")
*/
void fwUnitTestComponentTestDiscovery_getTestCaseRoutines(const dyn_string testScriptPathNames, mapping &scriptsTestRoutines, string sSelectionMode="Standard|Background")
{
  dyn_string testRoutines;
  for (int i = 1; i <= dynlen(testScriptPathNames); i++)
  {    
    if(fwUnitTestComponentTestDiscovery_getTestRoutinesFromScript(testScriptPathNames[i], testRoutines, sSelectionMode))
    {
      scriptsTestRoutines[testScriptPathNames[i]] = testRoutines;
      dynClear(testRoutines); 
      //DebugTN(scriptsTestRoutines[testScriptPathNames[i]]);   
    }             
  }   
}

//------------------------------------------------------------------------------------------------

/** 
  @par Description: Finds all test routines in the given file. 
  
  @par Usage
	  Public

  @param[in]  testScriptPathNames string 
  @param[out] testRoutines dyn_string, The test routines found within the testScript
  @param[in]  sSelectionMode Test selection mode (Standard, Interactive, Background, combined by "|")
*/
bool fwUnitTestComponentTestDiscovery_getTestRoutinesFromScript(const string testScriptPathName, dyn_string &testRoutines, string sSelectionMode="Standard|Background")
{
  dyn_string testFunctionPattern = fwUnitTestComponentTestDiscovery_getTestPatterns(sSelectionMode);
  bool bFoundRoutines = fwUnitTestComponentTestDiscovery_getRoutinesThatMatchPattern(testScriptPathName, testFunctionPattern, testRoutines);
  
  // remove any setup/teardown routines from the list of test cases.
  for (int i = 1; i <= dynlen(testRoutines); i++)
  {
    if (patternMatch("*test*setup*", testRoutines[i]) || (patternMatch("*test*teardown*", testRoutines[i])))
    {
      dynRemove(testRoutines, i);
      i--;
    }
  } 
  return bFoundRoutines;
}

//------------------------------------------------------------------------------------------------
/**
 * Gets the test patterns depending on the seleciton mode.
 * @param [in]  sSelectionMode Test selection mode (Standard, Interactive, Background, combined by "|")
 */
private dyn_string fwUnitTestComponentTestDiscovery_getTestPatterns(string sSelectionMode) {
  dyn_string dsModes = strsplit(sSelectionMode, "|");
  
  dyn_string testFunctionPattern;
  if (dynContains(dsModes, FWUNIT_TEST_STANDARD_MODE) > 0) {
    dynAppend(testFunctionPattern, "*void test*()*");
  }
  if (dynContains(dsModes, FWUNIT_TEST_INTERACTIVE_MODE) > 0) {
    dynAppend(testFunctionPattern, "*void interactiveTest*()*");
  }
  if (dynContains(dsModes, FWUNIT_TEST_BACKGROUND_MODE) > 0) {
    dynAppend(testFunctionPattern, "*void backgroundTest*()*");
  }
  
  return testFunctionPattern;
}
//------------------------------------------------------------------------------------------------

/** 
  @par Description: Gets the suiteSetup routine belonging to the given test script.
  
  @par Usage
	  Public

  @param[in]  testScriptPathName string, 
  @param[out] suiteSetupRoutine string, 
*/
bool fwUnitTestComponentTestDiscovery_getSuiteSetupRoutineFromScript(const string testScriptPathName, string &suiteSetupRoutine)
{
  dyn_string testRoutines;
  dyn_string testFunctionPattern = makeDynString("*void test*setupSuite()*");
  bool bFoundRoutines = fwUnitTestComponentTestDiscovery_getRoutinesThatMatchPattern(testScriptPathName, testFunctionPattern, testRoutines);
  if (bFoundRoutines)
  {
    suiteSetupRoutine = testRoutines[1]; 
  }
  return bFoundRoutines;
}

//------------------------------------------------------------------------------------------------

/** 
  @par Description: Gets the suiteTeardown routine belonging to the given test script.
  
  @par Usage
	  Public

  @param[in]  testScriptPathName string, 
  @param[out] suiteTeardownRoutine string, 
*/
bool fwUnitTestComponentTestDiscovery_getSuiteTeardownRoutineFromScript(const string testScriptPathName, string &suiteTeardownRoutine)
{
  dyn_string testRoutines;
  dyn_string testFunctionPattern = makeDynString("*void test*teardownSuite()*");
  bool bFoundRoutines = fwUnitTestComponentTestDiscovery_getRoutinesThatMatchPattern(testScriptPathName, testFunctionPattern, testRoutines);
  if (bFoundRoutines)
  {
    suiteTeardownRoutine = testRoutines[1];
  }
  return bFoundRoutines; 
}

//------------------------------------------------------------------------------------------------

/** 
  @par Description: Gets the Setup routine belonging to the given test script.
  
  @par Usage
	  Public

  @param[in]  testScriptPathName string, 
  @param[out] caseSetupRoutine string, 
*/
bool fwUnitTestComponentTestDiscovery_getCaseSetupRoutineFromScript(const string testScriptPathName, string &caseSetupRoutine)
{
  dyn_string testRoutines;
  dyn_string testFunctionPattern = makeDynString("*void test*setup()*");
  bool bFoundRoutines = fwUnitTestComponentTestDiscovery_getRoutinesThatMatchPattern(testScriptPathName, testFunctionPattern, testRoutines);
  if (bFoundRoutines)
  {
    caseSetupRoutine = testRoutines[1];
  }
  return bFoundRoutines; 
}

//------------------------------------------------------------------------------------------------

/** 
  @par Description: Gets the Teardown routine belonging to the given test script.
  
  @par Usage
	  Public

  @param[in]  testScriptPathName string, 
  @param[out] caseTeardownRoutine string, 
*/
bool fwUnitTestComponentTestDiscovery_getCaseTeardownRoutineFromScript(const string testScriptPathName, string &caseTeardownRoutine)
{
  dyn_string testRoutines;
  dyn_string testFunctionPattern = makeDynString("*void test*teardown()*");
  bool bFoundRoutines = fwUnitTestComponentTestDiscovery_getRoutinesThatMatchPattern(testScriptPathName, testFunctionPattern, testRoutines);
  if (bFoundRoutines)
  {
    caseTeardownRoutine = testRoutines[1];
  }
  return bFoundRoutines; 
}

//------------------------------------------------------------------------------------------------

/** 
  @par Description: Gets a list of all routines within the given script, that match any of the given patterns.
  
  @par Usage
	  Public

  @param[in]  testScriptPathName string, Script to be searched.
  @param[in]  patterns dyn_string 
  @param[out] testRoutines dyn_string, List of routines that matches the given patterns.
*/
private bool fwUnitTestComponentTestDiscovery_getRoutinesThatMatchPattern(const string testScriptPathName, const dyn_string patterns, dyn_string &testRoutines)
{
  bool foundMatchingFunction = false;
  string line;    
  string fileAsString;
  fileToString(testScriptPathName, fileAsString);
  dyn_string fileAsArray = strsplit(fileAsString, "\n");
  dyn_string fileAsArrayNoComments;    
  
  fwUnitTestComponentRemoveComments_RemoveComments(fileAsArray, fileAsArrayNoComments);
  
  for (int i = 1; i <= dynlen(fileAsArrayNoComments); i++)  
  {
    line = fileAsArrayNoComments[i];
    if (fwUnitTestComponentTestDiscovery_multiPatternMatch(patterns, line))
    {
      line = strltrim(line);
      line = strrtrim(line);
      if (strpos(line, "void") == 0) { 
        line = substr(line, strlen("void"));
        line = strltrim(line);
      }
      line = strrtrim(line, " {");
      dynAppend(testRoutines, line);       
      foundMatchingFunction = true; 
    }
  }
  return foundMatchingFunction; 
}  
 
//------------------------------------------------------------------------------------------------

/** 
  @par Description: @see fwUnitTestComponentTestDiscovery_getRoutinesThatMatchPattern()
  
  @par Usage
	  Public
   
  @param[in] patterns dyn_string, 
  @param[in] line string, 
  
  @return bool, true if the line matches any of the given patterns. 
  
*/
private bool fwUnitTestComponentTestDiscovery_multiPatternMatch(const dyn_string patterns, const string line)
{
  for (int  i = 1; i <= dynlen(patterns); i++)
  {
    if (patternMatch(patterns[i], line))
    {
      return true;
    }
  }
  return false;
}

//------------------------------------------------------------------------------------------------
